﻿Imports System.Text
Imports System.Threading.Tasks

Namespace SplitAndMerge
  Public Class Parser
	Public Const START_ARG As Char = "("c
	Public Const END_ARG As Char = ")"c
	Public Const END_LINE As Char = ControlChars.Lf

	Private Class Cell
	  Friend Sub New(ByVal value As Double, ByVal action As Char)
		Me.Value = value
		Me.Action = action
	  End Sub

            Friend Property Value As Double
            Friend Property Action As Char
        End Class

	Public Shared Function process(ByVal data As String) As Double
	  ' Get rid of spaces and check parenthesis
	  Dim expression As String = preprocess(data)
            Dim from As Integer = 0

            Return loadAndCalculate(data, from, END_LINE)
        End Function

	Private Shared Function preprocess(ByVal data As String) As String
	  If String.IsNullOrEmpty(data) Then
		Throw New ArgumentException("Loaded empty data")
	  End If

	  Dim parentheses As Integer = 0
	  Dim result As New StringBuilder(data.Length)

            For i = 0 To data.Length - 1
                Dim ch As Char = data.Chars(i)
                Select Case ch
                    Case " "c, ControlChars.Tab, ControlChars.Lf
                        Continue For
                    Case END_ARG
                        parentheses -= 1
                    Case START_ARG
                        parentheses += 1
                End Select
                result.Append(ch)
            Next i

            If parentheses <> 0 Then
		Throw New ArgumentException("Uneven parenthesis")
	  End If

	  Return result.ToString()
	End Function

        Public Shared Function loadAndCalculate(ByVal data As String, ByRef from As Integer, Optional ByVal [to] As Char = END_LINE) As Double
            If from >= data.Length OrElse data.Chars(from) = [to] Then
                Throw New ArgumentException("Loaded invalid data: " & data)
            End If

            Dim listToMerge As New List(Of Cell)(16)
            Dim item As New StringBuilder

            Do
                Dim ch As Char = data.Chars(from)
                from += 1
                If stillCollecting(item.ToString(), ch, [to]) Then
                    item.Append(ch)
                    If from < data.Length AndAlso data.Chars(from) <> [to] Then
                        Continue Do
                    End If
                End If

                ' We are done getting the next token. The getValue() call below may
                ' recursively call loadAndCalculate(). This will happen if extracted
                ' item is a function or if the next item is starting with a START_ARG '('.
                Dim func As New ParserFunction(data, from, item.ToString(), ch)
                Dim value As Double = func.getValue(data, from)

                Dim action As Char = If(validAction(ch), ch, updateAction(data, from, ch, [to]))

                listToMerge.Add(New Cell(value, action))
                item.Clear()

            Loop While from < data.Length AndAlso data.Chars(from) <> [to]

            If from < data.Length AndAlso (data.Chars(from) = END_ARG OrElse data.Chars(from) = [to]) Then
                from += 1
            End If

            Dim baseCell As Cell = listToMerge(0)
            Dim index As Integer = 1

            Return merge(baseCell, index, listToMerge)
        End Function

        Private Shared Function stillCollecting(ByVal item As String, ByVal ch As Char, ByVal [to] As Char) As Boolean
	  ' Stop collecting if either got END_ARG ')' or to char, e.g. ','.
	  Dim stopCollecting As Char = If([to] = END_ARG OrElse [to] = END_LINE, END_ARG, [to])
	  Return (item.Length = 0 AndAlso (ch = "-"c OrElse ch = END_ARG)) OrElse Not(validAction(ch) OrElse ch = START_ARG OrElse ch = stopCollecting)
	End Function

	Private Shared Function validAction(ByVal ch As Char) As Boolean
	  Return ch = "*"c OrElse ch = "/"c OrElse ch = "+"c OrElse ch = "-"c OrElse ch = "^"c
	End Function

        Private Shared Function updateAction(ByVal item As String, ByRef from As Integer, ByVal ch As Char, ByVal [to] As Char) As Char
            If from >= item.Length OrElse item.Chars(from) = END_ARG OrElse item.Chars(from) = [to] Then
                Return END_ARG
            End If

            Dim index As Integer = from
            Dim res As Char = ch
            Do While (Not validAction(res)) AndAlso index < item.Length
                res = item.Chars(index)
                index += 1
            Loop

            from = If(validAction(res), index, If(index > from, index - 1, from))
            Return res
        End Function

        ' From outside this function is called with mergeOneOnly = false.
        ' It also calls itself recursively with mergeOneOnly = true, meaning
        ' that it will return after only one merge.
        Private Shared Function merge(ByVal current As Cell, ByRef index As Integer, ByVal listToMerge As List(Of Cell), Optional ByVal mergeOneOnly As Boolean = False) As Double
	  Do While index < listToMerge.Count
		Dim [next] As Cell = listToMerge(index)
		index += 1

		Do While Not canMergeCells(current, [next])
		  ' next cells first. E.g. if we have 1+2*3, we first merge next
		  ' cells, i.e. 2*3, getting 6, and then we can merge 1+6.
		  merge([next], index, listToMerge, True) ' mergeOneOnly
		Loop
		mergeCells(current, [next])
		If mergeOneOnly Then
		  Return current.Value
		End If
	  Loop

	  Return current.Value
	End Function

	Private Shared Sub mergeCells(ByVal leftCell As Cell, ByVal rightCell As Cell)
	  Select Case leftCell.Action
		Case "^"c
			leftCell.Value = Math.Pow(leftCell.Value, rightCell.Value)
		Case "*"c
			leftCell.Value *= rightCell.Value
		Case "/"c
		  If rightCell.Value = 0 Then
			Throw New ArgumentException("Division by zero")
		  End If
		  leftCell.Value /= rightCell.Value
		Case "+"c
			leftCell.Value += rightCell.Value
		Case "-"c
			leftCell.Value -= rightCell.Value
	  End Select
	  leftCell.Action = rightCell.Action
	End Sub

	Private Shared Function canMergeCells(ByVal leftCell As Cell, ByVal rightCell As Cell) As Boolean
	  Return getPriority(leftCell.Action) >= getPriority(rightCell.Action)
	End Function

	Private Shared Function getPriority(ByVal action As Char) As Integer
	  Select Case action
		Case "^"c
			Return 4
		Case "*"c, "/"c
			Return 3
		Case "+"c, "-"c
			Return 2
	  End Select
	  Return 0
	End Function
  End Class

  Public Class ParserFunction
	Public Sub New()
	  m_impl = Me
	End Sub

        ' A "virtual" Constructor
        Friend Sub New(ByVal data As String, ByRef from As Integer, ByVal item As String, ByVal ch As Char)
            If item.Length = 0 AndAlso ch = Parser.START_ARG Then
                ' There is no function, just an expression in parentheses
                m_impl = s_idFunction
                Exit Sub
            End If

            If m_functions.TryGetValue(item, m_impl) Then
                ' Function exists and is registered (e.g. pi, exp, etc.)
                Exit Sub
            End If

            ' Function not found, will try to parse this as a number.
            s_strtodFunction.Item = item
            m_impl = s_strtodFunction
        End Sub

        Public Shared Sub addFunction(ByVal name As String, ByVal [function] As ParserFunction)
            m_functions(name) = [function]
        End Sub

        Public Function getValue(ByVal data As String, ByRef from As Integer) As Double
            Return m_impl.evaluate(data, from)
        End Function

        Protected Overridable Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            ' The real implementation will be in the derived classes.
            Return 0
        End Function

        Private m_impl As ParserFunction
        Private Shared m_functions As New Dictionary(Of String, ParserFunction)

        Private Shared s_strtodFunction As New StrtodFunction
        Private Shared s_idFunction As New IdentityFunction
    End Class

  Friend Class StrtodFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Dim num As Double
            If Not Double.TryParse(Item, num) Then
                Throw New ArgumentException("Could not parse token [" & Item & "]")
            End If
            Return num
        End Function
        Private privateItem As String
        Public Property Item As String
            Private Get
                Return privateItem
            End Get
            Set(ByVal value As String)
                privateItem = value
            End Set
        End Property
    End Class

  Friend Class IdentityFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Return Parser.loadAndCalculate(data, from, Parser.END_ARG)
        End Function
    End Class

  Friend Class PiFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Return 3.1415926535897931
        End Function
    End Class
  Friend Class ExpFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Dim arg As Double = Parser.loadAndCalculate(data, from, Parser.END_ARG)
            Return Math.Exp(arg)
        End Function
    End Class
  Friend Class PowFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Dim arg1 As Double = Parser.loadAndCalculate(data, from, ","c)
            Dim arg2 As Double = Parser.loadAndCalculate(data, from, Parser.END_ARG)

            Return Math.Pow(arg1, arg2)
        End Function
    End Class
  Friend Class SinFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Dim arg As Double = Parser.loadAndCalculate(data, from, Parser.END_ARG)
            Return Math.Sin(arg)
        End Function
    End Class
  Friend Class SqrtFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Dim arg As Double = Parser.loadAndCalculate(data, from, Parser.END_ARG)
            Return Math.Sqrt(arg)
        End Function
    End Class
  Friend Class AbsFunction
	  Inherits ParserFunction
        Protected Overrides Function evaluate(ByVal data As String, ByRef from As Integer) As Double
            Dim arg As Double = Parser.loadAndCalculate(data, from, Parser.END_ARG)
            Return Math.Abs(arg)
        End Function
    End Class

  Friend Class Program
	Private Shared Sub calculate(ByVal expr As String, ByVal expected As Double)
	  Dim result As Double = Parser.process(expr)

            Dim outcome As String = If(result = expected, "OK", "NOK " & expected.ToString)
            Console.WriteLine("{0} --> {1} ({2})", expr, result, outcome)
	End Sub

	Shared Sub Main(ByVal args() As String)
            ParserFunction.addFunction("pi", New PiFunction)
            ParserFunction.addFunction("exp", New ExpFunction)
            ParserFunction.addFunction("pow", New PowFunction)
            ParserFunction.addFunction("sin", New SinFunction)
            ParserFunction.addFunction("abs", New AbsFunction)
            ParserFunction.addFunction("sqrt", New SqrtFunction)

            calculate("(((-5.5)))", (((-5.5))))
	  calculate("1-2", 1 - 2)
	  calculate("(1-(2))", (1 - (2)))
	  calculate("3+2*6-1", 3 + 2 * 6 - 1)
	  calculate("3-2*6-1", 3 - 2 * 6 - 1)
	  calculate("1-2-3-(4-(5-(6-7)))", 1 - 2 - 3 - (4 - (5 - (6 - 7))))
	  calculate("2-3*sin(pi)", 2 - 3 * Math.Sin(Math.PI))
	  calculate("1-(exp(10*7-sqrt((1+1)*20*10)))", 1 - (Math.Exp(10 * 7 - Math.Sqrt((1 + 1) * 20 * 10))))
	  calculate("3-(5-6)-(2-(3-(1-2)))", 3 - (5 - 6) - (2 - (3 - (1 - 2))))
	  calculate("3-(5-6)-(2-(3-(1+2)))+2-(-1+7)*(9-2)/((16-3)-3)+15/2*5", 3 - (5 - 6) - (2 - (3 - (1 + 2))) + 2 - (-1 + 7) * (9 - 2) / ((16.0 - 3) - 3.0) + 15 / 2.0 * 5)
	  calculate("(-1+7)*(9-2)", (-1 + 7) * (9 - 2))
	  calculate("((16-3)-3)+15/2*5", ((16 - 3) - 3) + 15 / 2.0 * 5)
	  calculate("1+15/2*5", 1 + 15 / 2.0 * 5)
	  calculate("3-2/6-1", 3-2/6.0-1)
	  calculate("3*50-3*2^4*3", 3 * 50 - 3 * Math.Pow(2, 4) * 3)
	  calculate("5-1/2^2-3", 5 - 1 / Math.Pow(2, 2) - 3)
	  calculate("(((1/4/2-(8/2/3+5))))", (((1 / 4.0 / 2.0 - (8 / 2.0 / 3.0 + 5)))))
	  calculate("pow(2,3)", Math.Pow(2, 3))
	  calculate("abs(3*-50-2*3/4)/3*2", Math.Abs(3.0 * -50 - 2 * 3.0 / 4.0) / 3.0 * 2)

	  Console.ReadKey()
	End Sub
  End Class
End Namespace
